﻿using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Globalization;
using System.ComponentModel;
using System.Data.Linq;
using System.Web.DynamicData;
using System.Collections.Generic;
using System.Collections.Specialized;

namespace System.Web.Mvc {
    public abstract class DynamicDataController<T> : Controller {
        private DynamicMetaTable _metaTable;
        private DynamicViewData<T> _dynamicViewData;
        private DynamicFormErrorInfo _formErrorInfo;

        public DynamicDataController() {
            _metaTable = DynamicDatabase.TheDatabase.GetMetaTable(typeof(T));
        }

        [ControllerAction]
        public virtual void Index() {
            InvokeAction("list");
        }

        [ControllerAction]
        public virtual void List(string sort) {
            // Start with a full query
            var query = ListQuery;

            // Add a where clause for each foreign key specified on the query string
            foreach (string name in Request.QueryString) {
                var member = MetaTable.TryFindColumn(name) as DynamicMetaForeignKeyMember;
                if (member == null)
                    continue;

                query = DynamicMvcUtil.ApplyForeignKeyWhereClause(MetaTable, member, query, Request.QueryString[name]);
            }

            // Sort the query if needed
            if (!String.IsNullOrEmpty(sort)) {
                query = DynamicMvcUtil.ApplyOrderByClause(query, sort);
            }

            DynamicViewData.Data = query;

            RenderView("list");
        }

        [ControllerAction]
        public virtual void Show(string id) {
            DynamicViewData.Data = GetItemQuery(id);

            RenderView("show");
        }

        [ControllerAction]
        public virtual void New() {
            EnsureDynamicViewData();

            RenderView("new");
        }

        [ControllerAction]
        public virtual void Create() {

            // If it was cancelled, just go back to the 'index' view
            if (!String.IsNullOrEmpty(Request.Form["formSubmit"])) {
                object dataObject = Activator.CreateInstance(MetaTable.EntityType);

                FillEntityPropertiesFromRequest(dataObject, Request.QueryString);

                // If there are errors, go back to the New action
                if (!FillEntityPropertiesFromForm(dataObject)) {
                    RedirectToActionOrReturnPath("new");
                }

                ITable table = (ITable)MetaTable.DataContextProperty.GetValue(DataContext, null);
                table.InsertOnSubmit(dataObject);

                SubmitChanges("new" /*nextAction*/);

                TempData.SetStatusMessage("The item was successfully created!");
            }

            RedirectToActionOrReturnPath("list");
        }

        [ControllerAction]
        public virtual void Edit(string id) {
            EnsureDynamicViewData();

            DynamicViewData.Data = GetItemQuery(id);

            RenderView("edit");
        }

        [ControllerAction]
        public virtual void Update(string id) {
            // If the change was submitted
            if (!String.IsNullOrEmpty(Request.Form["formSubmit"])) {
                object dataObject = GetItem(id);

                // Make sure we have write access to the object
                DynamicSecurityUtility.EnsureWriteAccess(dataObject);

                // If there are errors, go back to the Edit action
                if (!FillEntityPropertiesFromForm(dataObject)) {
                    RedirectToActionWithID("edit", id);
                }

                SubmitChanges("edit" /*nextAction*/);

                TempData.SetStatusMessage("The update was successful!");
            }

            RedirectToActionWithID("show", id);
        }

        [ControllerAction]
        public virtual void Destroy(string id) {
            object dataObject = GetItem(id);

            // Make sure we have write access to the object
            DynamicSecurityUtility.EnsureWriteAccess(dataObject);

            ITable table = (ITable)MetaTable.DataContextProperty.GetValue(DataContext, null);
            table.DeleteOnSubmit(dataObject);
            DataContext.SubmitChanges();

            TempData.SetStatusMessage("The item was successfully deleted!");

            RedirectToActionOrReturnPath("list");
        }

        // Go to the return path if there is one.  Otherwise, go to the passed in action
        protected void RedirectToActionOrReturnPath(string actionName) {
            string returnPath = Request.QueryString[DynamicMvcUtil.ReturnPathQueryStringKey];

            if (!String.IsNullOrEmpty(returnPath))
                Response.Redirect(returnPath);
            else
                RedirectToAction(actionName);
        }

        protected bool FillEntityPropertiesFromForm(object dataObject) {
            return FillEntityPropertiesFromRequest(dataObject, Request.Form);
        }

        // Fill in the properties of an entity instance from the fields found in the form
        protected bool FillEntityPropertiesFromRequest(object dataObject, NameValueCollection parameters) {

            // Go through all the entries in the posted form
            foreach (string propName in parameters) {
                DynamicMetaMember member = MetaTable.TryFindColumn(propName);
                if (member == null)
                    continue;

                // Get the posted value for that property
                string propValue = parameters[propName];

                // Convert empty fields to null if we're supposed to
                if (propValue == String.Empty && member.ConvertEmptyStringToNull)
                    propValue = null;

                // Fail if a required field is missing
                if (propValue == null && member.IsRequired) {
                    string error = member.RequiredErrorMessage;
                    if (String.IsNullOrEmpty(error)) {
                        error = String.Format("The field '{0}' is required", member.Name);
                    }
                    AddFormFieldError(member, error);
                    continue;
                }

                // Check if it's a foreign key
                var fkMember = member as DynamicMetaForeignKeyMember;

                object value = null;
                if (fkMember != null) {
                    // For foreign keys, we need to get the parent entity in order to assign it
                    IQueryable parentEntityQuery = DynamicMvcUtil.ApplyPrimaryKeyWhereClause(
                        fkMember.ParentMetaTable,
                        fkMember.ParentMetaTable.GetQuery(DataContext),
                        propValue);
                    value = parentEntityQuery.SingleUntyped();
                }
                else if (propValue != null) {

                    // Convert the posted value to the correct type
                    TypeConverter converter = TypeDescriptor.GetConverter(member.Type);
                    try {
                        value = converter.ConvertFromString(propValue);
                    }
                    catch (Exception e) {
                        AddFormFieldError(member, e.Message);
                        continue;
                    }
                }

                if (propValue != null) {
                    // TODO: Just handle Range for now for demonstration purpose.  Later, this
                    // will be done through attribute self validation
                    var range = (RangeAttribute)member.GetMetaDataAttribute(typeof(RangeAttribute));
                    if (range != null) {
                        if ((int)value < (int)range.Minimum || (int)value > (int)range.Maximum) {
                            AddFormFieldError(member, String.Format("Must be between {0} and {1}", (int)range.Minimum, (int)range.Maximum));
                            continue;
                        }
                    }
                }

                try {
                    // Set the property value in the entity
                    member.MemberPropertyInfo.SetValue(dataObject, value, null);
                }
                catch (TargetInvocationException e) {
                    AddFormFieldError(member, e.InnerException.Message);
                    continue;
                }
            }

            return _formErrorInfo == null;
        }

        private void SubmitChanges(string nextAction) {
            try {
                DataContext.SubmitChanges();
            }
            catch (Exception e) {
                AddEntityError(e.Message);
                RedirectToActionOrReturnPath(nextAction);
            }
        }

        private void AddFormFieldError(DynamicMetaMember member, string message) {
            EnsureForErrorInfo();
            _formErrorInfo[member.Name].ErrorMessages.Add(message);
        }

        private void AddEntityError(string message) {
            EnsureForErrorInfo();
            _formErrorInfo.ErrorMessages.Add(message);
        }

        private void EnsureForErrorInfo() {
            if (_formErrorInfo == null) {
                _formErrorInfo = new DynamicFormErrorInfo(Request.Form);
                TempData.SetFormErrorInfo(_formErrorInfo);
            }
        }

        protected virtual IQueryable<T> GetItemQuery(string id) {
            var query = ViewQuery;
            if (!String.IsNullOrEmpty(id)) {
                query = DynamicMvcUtil.ApplyPrimaryKeyWhereClause(MetaTable, query, id);
            }

            return query;
        }

        protected T GetItem(string id) {
            return GetItemQuery(id).Single();
        }

        protected DynamicViewData<T> DynamicViewData {
            get {
                EnsureDynamicViewData();

                return _dynamicViewData;
            }
        }

        private void EnsureDynamicViewData() {
            if (_dynamicViewData == null) {
                _dynamicViewData = new DynamicViewData<T>(this);
                _dynamicViewData.MetaTable = MetaTable;
            }
        }

        protected virtual IQueryable<T> ListQuery {
            get {
                return (IQueryable<T>)MetaTable.GetQuery(DataContext);
            }
        }

        protected virtual IQueryable<T> ViewQuery {
            get {
                return (IQueryable<T>)MetaTable.GetQuery(DataContext);
            }
        }

        // REVIEW: need to use an abstraction instead of hardcoding to DLINQ
        public DataContext DataContext {
            get {
                return HttpContext.GetDataContext();
            }
        }

        protected DynamicMetaTable MetaTable {
            get {
                return _metaTable;
            }
        }

        // REVIEW: this should be on the base Controller class
        protected void RedirectToActionWithID(string actionName, string ID) {
            RedirectToAction(new { Action = actionName, ID = ID });
        }
    }

    public class DynamicFormErrorInfo {

        internal DynamicFormErrorInfo(NameValueCollection form) {
            ErrorMessages = new List<string>();

            FieldErrors = new Dictionary<string, DynamicFormFieldErrorInfo>(StringComparer.OrdinalIgnoreCase);

            // Copy all the values from the form into our field info objects
            foreach (string key in form.AllKeys) {
                FieldErrors[key] = new DynamicFormFieldErrorInfo(key, form[key]);
            }
        }

        // List of error messages that relate to the entity as a whole (and not any individual field)
        public IList<string> ErrorMessages { get; private set; }

        // The key is the name of the property and the value the associated data
        public IDictionary<string, DynamicFormFieldErrorInfo> FieldErrors { get; private set; }


        // The key is the name of the property and the value the associated data
        public DynamicFormFieldErrorInfo this[string fieldName] { get { return FieldErrors[fieldName]; } }
    }

    public class DynamicFormFieldErrorInfo {

        internal DynamicFormFieldErrorInfo(string fieldName, string previousValue) {
            FieldName = fieldName;
            PreviousValue = previousValue;
            ErrorMessages = new List<string>();
        }

        public string FieldName { get; private set; }

        // List of error messages for that field
        public IList<string> ErrorMessages { get; private set; }

        // Previous value that was displayed in that field
        public string PreviousValue { get; private set; }
    }
}
