﻿using System.Web.UI;
using System.CodeDom;
using System.Collections;
using System.Web.Compilation;
using System.Web.Mvc;

namespace System.Web.DynamicData.Controls {
    public class DynamicDataHelperBuilder : ControlBuilder {

        private const string DynamicViewDataPropName = "DVD";
        private IDictionary _attribs;
        private string _id;

        public override void Init(TemplateParser parser, ControlBuilder parentBuilder, Type type,
            string tagName, string id, IDictionary attribs) {

            if (!String.IsNullOrEmpty(id)) {
                _id = id;
            }
            else {
                _id = DynamicViewDataPropName;
            }

            // Pass null for the ID since we want to use it differently
            base.Init(parser, parentBuilder, type, tagName, null, attribs);

            _attribs = attribs;
        }

        public override void ProcessGeneratedCode(CodeCompileUnit codeCompileUnit, CodeTypeDeclaration baseType,
            CodeTypeDeclaration derivedType, CodeMemberMethod buildMethod, CodeMemberMethod dataBindingMethod) {

            string DynamicViewDataFieldName = "__" + _id;

            string entityTypeNamesAttrib = (string)_attribs["Entities"];
            if (String.IsNullOrEmpty(entityTypeNamesAttrib))
                return;

            string[] entityNames = entityTypeNamesAttrib.Split(',');

            DynamicMetaTable primaryTable = DynamicDatabase.TheDatabase.GetMetaTable(entityNames[0]);

            // e.g. typeof(DynamicViewData<Comment>)
            var viewDataTypeRef = new CodeTypeReference(typeof(DynamicViewData));
            viewDataTypeRef.TypeArguments.Add(new CodeTypeReference(primaryTable.EntityType));

            // e.g. DynamicViewData<Comment> __dvd;
            var dynamicViewDataField = new CodeMemberField(viewDataTypeRef, DynamicViewDataFieldName);
            derivedType.Members.Add(dynamicViewDataField);

            // e.g. this.__dvd
            var viewDataFieldRef = new CodeFieldReferenceExpression(new CodeThisReferenceExpression(), DynamicViewDataFieldName);

            // e.g. DynamicViewData<Comment> DVD {
            //  get {
            //      if (__dvd == null) {
            //          __dvd = DynamicViewData<MvcBlog.Models.Comment>.GetDynamicViewData(this);
            //      }
            //      return __dvd;
            //  }
            //
            var prop = new CodeMemberProperty();
            derivedType.Members.Add(prop);
            prop.Name = _id;
            prop.Type = viewDataTypeRef;
            var testViewDataField = new CodeConditionStatement();
            prop.GetStatements.Add(testViewDataField);
            testViewDataField.Condition = new CodeBinaryOperatorExpression(
                viewDataFieldRef, CodeBinaryOperatorType.ValueEquality, new CodePrimitiveExpression(null));
            testViewDataField.TrueStatements.Add(new CodeAssignStatement(
                viewDataFieldRef,
                new CodeMethodInvokeExpression(
                    new CodeTypeReferenceExpression(viewDataTypeRef),
                    "GetDynamicViewData",
                    new CodeThisReferenceExpression())));
            prop.GetStatements.Add(new CodeMethodReturnStatement(viewDataFieldRef));

            // e.g. this.DVD
            var dynamicViewDataRef = new CodePropertyReferenceExpression(new CodeThisReferenceExpression(), _id);

            // e.g. this.DVD.Render()
            var renderMethodRef = new CodeMethodReferenceExpression(dynamicViewDataRef, "Render");
            var renderEditMethodRef = new CodeMethodReferenceExpression(dynamicViewDataRef, "RenderEdit");
            var renderInsertMethodRef = new CodeMethodReferenceExpression(dynamicViewDataRef, "RenderInsert");

            foreach (string entityName in entityNames) {
                DynamicMetaTable metaTable = DynamicDatabase.TheDatabase.GetMetaTable(entityName);
                var entityType = metaTable.EntityType;

                foreach (var member in metaTable.MetaMembers) {
                    derivedType.Members.Add(CreateMethod(entityType, member, renderMethodRef, ""));
                    derivedType.Members.Add(CreateMethod(entityType, member, renderEditMethodRef, "_Edit"));
                    derivedType.Members.Add(CreateInsertMethod(entityType, member, renderInsertMethodRef, "_Insert"));
                }
            }
        }

        private CodeMemberMethod CreateMethod(Type entityType, DynamicMetaMember member,
            CodeMethodReferenceExpression renderMethodRef, string methodNameSuffix) {
            // e.g. private void Comment_Title(MvcBlog.Models.Comment comment) {
            //      this.DVD.Render(comment, "Title");
            // }
            var method = new CodeMemberMethod();
            method.Name = entityType.Name + "_" + member.Name + methodNameSuffix;
            string paramName = entityType.Name.ToLowerInvariant();
            method.Parameters.Add(new CodeParameterDeclarationExpression(entityType, paramName));

            // e.g. this.DVD.Render(comment, "Title");
            method.Statements.Add(new CodeMethodInvokeExpression(
                renderMethodRef,
                new CodeVariableReferenceExpression(paramName),
                new CodePrimitiveExpression(member.Name)));

            return method;
        }

        private CodeMemberMethod CreateInsertMethod(Type entityType, DynamicMetaMember member,
            CodeMethodReferenceExpression renderMethodRef, string methodNameSuffix) {
            // e.g. private void Comment_Title_Insert() {
            //      this.DVD.Render_Insert(typeof(comment), "Title");
            // }
            var method = new CodeMemberMethod();
            method.Name = entityType.Name + "_" + member.Name + methodNameSuffix;

            // e.g. this.DVD.Render_Insert(entityType, "Title");
            method.Statements.Add(new CodeMethodInvokeExpression(
                renderMethodRef,
                new CodeTypeOfExpression(entityType),
                new CodePrimitiveExpression(member.Name)));

            return method;
        }
    }

    [ControlBuilderAttribute(typeof(DynamicDataHelperBuilder))]
    public class DynamicDataHelper : Control {
        public string Entities { get; set; }
    }
}
