﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Web.Mvc;
using System.Web.UI;
using System.Web.DynamicData;
using System.Data.Linq;
using System.ComponentModel;

namespace System.Web.Mvc {

    public abstract class DynamicViewData {
        internal const string ViewDataSlot = "__DynamicViewData";
        private ViewPage _viewPage;
        private HtmlTextWriter _writer;
        private DynamicFormErrorInfo _formErrorInfo;

        public static DynamicViewData GetDynamicViewData(ViewPage viewPage) {
            var dynamicViewData = (DynamicViewData)viewPage.ViewData[ViewDataSlot];

            dynamicViewData.SetViewPage(viewPage);

            return dynamicViewData;
        }

        protected DynamicViewData(Controller controller) {
            controller.ViewData[ViewDataSlot] = this;
        }

        internal void SetViewPage(ViewPage viewPage) {
            _viewPage = viewPage;

            // Get the previous request's error info if any
            _formErrorInfo = viewPage.ViewContext.TempData.GetFormErrorInfo();
        }

        private HtmlTextWriter Writer {
            get {
                if (_writer == null) {
                    _writer = new HtmlTextWriter(_viewPage.Response.Output);
                }

                return _writer;
            }
        }

        public DataContext DataContext {
            get {
                return _viewPage.ViewContext.HttpContext.GetDataContext();
            }
        }

        public DynamicMetaTable MetaTable { get; set; }
        public IQueryable Data { get; set; }

        public object DataItem {
            get {
                return Data.SingleUntyped();
            }
        }

        public IEnumerable<DynamicMetaMember> DisplayMetaMembers {
            get {
                foreach (var member in MetaTable.MetaMembers) {
                    if (!member.ShouldBeDisplayedInAutoMode)
                        continue;

                    yield return member;
                }
            }
        }

        public IEnumerable<DynamicMetaMember> DisplayShortMetaMembers {
            get {
                foreach (var member in MetaTable.MetaMembers) {
                    if (!member.ShouldBeDisplayedInAutoMode || member.IsLongString)
                        continue;

                    yield return member;
                }
            }
        }

        public DynamicFormErrorInfo FormErrorInfo {
            get { return _formErrorInfo; }
        }

        // REVIEW: TEMPORARY: The MVC framework needs to provide this.
        private string MakeActionUrl(string controller, string action, string id) {

            object values;
            if (!String.IsNullOrEmpty(controller)) {
                values = new { Controller = controller, Action = action, ID = id };
            }
            else {
                values = new { Action = action, ID = id };
            }

            return _viewPage.Url.Action(values);
        }

        public string MakeUrl(string action, object entityInstance, bool includeReturnPath) {
            Type entityType = entityInstance.GetType();
            var metaTable = DynamicDatabase.TheDatabase.GetMetaTable(entityType);

            // REVIEW: temporary! need to deal with multiple PK case!
            var member = metaTable.IdentityMetaMembers[0];
            object value = DataBinder.GetPropertyValue(entityInstance, member.Name);
            string id = value.ToString();

            string url = MakeActionUrl(entityType.Name, action, id);

            if (includeReturnPath) {
                url += "?" + DynamicMvcUtil.ReturnPathQueryStringKey + "=" + _viewPage.Request.Path;
            }

            return url;
        }

        public string MakeUrl(string action, object entityInstance) {
            return MakeUrl(action, entityInstance, false /*includeReturnPath*/);
        }

        public string MakeUrl(string actionName, string controllerName, object values) {
            string url = _viewPage.Url.Action(actionName, controllerName);

            bool first = true;
            foreach (var prop in GetProperties(values)) {
                if (first) {
                    url += "?";
                    first = false;
                }
                else {
                    url += "&";
                }

                url += prop.Name;
                url += "=";

                // If the value is a known entity and we can get a primary key string from it, use it
                string value = DynamicMvcUtil.GetEntityKeyString(prop.Value);
                if (value == null)
                    value = prop.Value.ToString();

                url += value;
            }

            // REVIEW: should use some query string building helpers
            url += first ? "?" : "&";

            url += DynamicMvcUtil.ReturnPathQueryStringKey + "=" + _viewPage.Request.Path;

            return url;
        }

        // REVIEW: copied from RouteCollection.  Remove when we merge.
        private static IEnumerable<PropertyValue> GetProperties(object o) {
            if (o != null) {
                PropertyDescriptorCollection props = TypeDescriptor.GetProperties(o);
                foreach (PropertyDescriptor prop in props) {
                    object val = prop.GetValue(o);
                    yield return new PropertyValue { Name = prop.Name, Value = val };
                }
            }
        }
        // REVIEW: copied from RouteCollection.  Remove when we merge.
        private sealed class PropertyValue {
            public string Name;
            public object Value;
        }

        private void RenderField(object entityInstance, Type entityType, string fieldName, DynamicTemplateMode templateMode) {

            var metaTable = DynamicDatabase.TheDatabase.GetMetaTable(entityType);
            var metaMember = metaTable.FindColumn(fieldName);

            string fieldTemplatePath = FieldTemplateManager.GetFieldTemplatePath(
                null, metaMember, templateMode);

            if (fieldTemplatePath == null)
                return;

            var uc = (MvcFieldTemplateUserControlBase)_viewPage.LoadControl(fieldTemplatePath);

            uc.DynamicViewData = this;
            uc.MetaMember = metaMember;
            uc.EntityInstance = entityInstance;

            uc.RenderControl(Writer);
        }

        // REVIEW: experimental render method using linq expression trees.  The entity and property
        // name are parsed from the tree and then rendered.  It gives nice strong typing, but is
        // probably very inefficient.
        public delegate object Property();
        public void Render(Expression<Property> property) {
            var propertyAccess = (MemberExpression)property.Body;

            var lambda = Expression.Lambda<Func<object>>(propertyAccess.Expression);
            object entityInstance = lambda.Compile()();

            Render(entityInstance, propertyAccess.Member.Name);
        }

        public void Render(object entityInstance, string fieldName) {
            RenderField(entityInstance, entityInstance.GetType(), fieldName, DynamicTemplateMode.ItemTemplate);
        }

        public void RenderEdit(object entityInstance, string fieldName) {
            RenderField(entityInstance, entityInstance.GetType(), fieldName, DynamicTemplateMode.EditItemTemplate);
        }

        public void RenderInsert(Type entityType, string fieldName) {
            RenderField(null /*entityInstance*/, entityType, fieldName, DynamicTemplateMode.InsertItemTemplate);
        }
    }

    public class DynamicViewData<T> : DynamicViewData {
        private IQueryable<T> _data;

        public new static DynamicViewData<T> GetDynamicViewData(ViewPage viewPage) {
            var dynamicViewData = (DynamicViewData<T>)viewPage.ViewData[ViewDataSlot];

            dynamicViewData.SetViewPage(viewPage);

            return dynamicViewData;
        }

        public DynamicViewData(Controller controller) : base(controller) { }

        public new IQueryable<T> Data {
            get {
                return _data;
            }
            set {
                base.Data = value;
                _data = value;
            }
        }

        public new T DataItem {
            get {
                return Data.Single();
            }
        }
    }
}
