﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Web.DynamicData;

namespace Microsoft.Web.DynamicData.Metadata {
    internal class MetadataAttributeConverter {
        private static MetadataAttributeConverter s_instance;
        private List<IAttributeConverter> _converters = new List<IAttributeConverter>();

        private MetadataAttributeConverter() {
            // look through all nested types
            foreach (Type converterType in GetType().GetNestedTypes(BindingFlags.NonPublic)) {
                // for each concrete type that derives from IAttributeConverter, add it to the converter collecion
                if (!converterType.IsAbstract && typeof(IAttributeConverter).IsAssignableFrom(converterType)) {
                    ConstructorInfo constructor = converterType.GetConstructor(new Type[0]);
                    AddConverter((IAttributeConverter)constructor.Invoke(new object[0]));
                }
            }
        }

        public static MetadataAttributeConverter Instance {
            get {
                if (s_instance == null) {
                    s_instance = new MetadataAttributeConverter();
                }
                return s_instance;
            }
        }

        private void AddConverter(IAttributeConverter converter) {
            _converters.Add(converter);
        }

        public System.Web.DynamicData.BaseMetadataAttribute ConvertFromNewToOld(string propertyName, BaseMetadataAttribute attribute) {
            if (null == propertyName) {
                throw new ArgumentNullException("propertyName");
            }

            if (null == attribute) {
                throw new ArgumentNullException("attribute");
            }

            // find the appropriate converter for this type
            Type attributeType = attribute.GetType();
            foreach (IAttributeConverter converter in _converters) {
                if (converter.ConvertFromType == attributeType) {
                    return converter.Convert(propertyName, attribute);
                }
            }
            throw new InvalidOperationException(String.Format("No converter found for type {0}", attribute.GetType().FullName));
        }

        private interface IAttributeConverter {
            System.Web.DynamicData.BaseMetadataAttribute Convert(string propertyName, BaseMetadataAttribute attribute);
            Type ConvertFromType { get; }
            Type ConvertToType { get; }
        }

        private abstract class AttributeConverter<TNew, TOld> : IAttributeConverter
            where TNew : BaseMetadataAttribute
            where TOld : System.Web.DynamicData.BaseMetadataAttribute {

            public System.Web.DynamicData.BaseMetadataAttribute Convert(string propertyName, BaseMetadataAttribute attribute) {
                return Convert(propertyName, (TNew)attribute);
            }

            public abstract TOld Convert(string propertyName, TNew attribute);

            public Type ConvertFromType { get { return typeof(TNew); } }
            public Type ConvertToType { get { return typeof(TOld); } }

        }

        private class DescriptionConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.DescriptionAttribute, System.Web.DynamicData.DescriptionAttribute> {
            public override System.Web.DynamicData.DescriptionAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.DescriptionAttribute attribute) {
                return new System.Web.DynamicData.DescriptionAttribute(propertyName, attribute.Description);
            }
        }

        private class DisplayFormatConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.DisplayFormatAttribute, System.Web.DynamicData.DisplayFormatAttribute> {
            public override System.Web.DynamicData.DisplayFormatAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.DisplayFormatAttribute attribute) {
                var old = new System.Web.DynamicData.DisplayFormatAttribute(propertyName);
                old.ConvertEmptyStringToNull = attribute.ConvertEmptyStringToNull;
                old.DataFormatString = attribute.DataFormatString;
                old.NullDisplayText = attribute.NullDisplayText;
                return old;
            }
        }

        private class RangeConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.RangeAttribute, System.Web.DynamicData.RangeAttribute> {
            public override System.Web.DynamicData.RangeAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.RangeAttribute attribute) {
                Type rangeType = attribute.Maximum.GetType();
                ConstructorInfo constructor = typeof(System.Web.DynamicData.RangeAttribute).GetConstructor(new Type[] { typeof(string), rangeType, rangeType });
                if (null != constructor) {
                    var result = (System.Web.DynamicData.RangeAttribute)constructor.Invoke(new object[] { propertyName, attribute.Minimum, attribute.Maximum });
                    result.ErrorMessage = attribute.ErrorMessage;
                    return result;
                }

                throw new InvalidOperationException();
            }
        }

        private class RenderHintConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.RenderHintAttribute, System.Web.DynamicData.RenderHintAttribute> {
            public override System.Web.DynamicData.RenderHintAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.RenderHintAttribute attribute) {
                return new System.Web.DynamicData.RenderHintAttribute(propertyName, attribute.RenderHint);
            }
        }

        private class RegexConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.RegexAttribute, System.Web.DynamicData.RegexAttribute> {
            public override System.Web.DynamicData.RegexAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.RegexAttribute attribute) {
                return new System.Web.DynamicData.RegexAttribute(propertyName, attribute.Regex) { ErrorMessage = attribute.ErrorMessage };
            }
        }

        private class RequiredConverter : AttributeConverter<Microsoft.Web.DynamicData.Metadata.RequiredAttribute, System.Web.DynamicData.RequiredAttribute> {
            public override System.Web.DynamicData.RequiredAttribute Convert(string propertyName, Microsoft.Web.DynamicData.Metadata.RequiredAttribute attribute) {
                return new System.Web.DynamicData.RequiredAttribute(propertyName) { ErrorMessage = attribute.ErrorMessage };
            }
        }
    }
}
