﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data.Linq;
using System.Reflection;

namespace Microsoft.Web.DynamicData.Metadata {
    public static class MetadataHandlerRegistration {
        private static Dictionary<Type, List<TypeDescriptionProvider>> s_registeredProviders = new Dictionary<Type, List<TypeDescriptionProvider>>();
        private static List<Type> s_registeredDataContexts = new List<Type>();
        
        public static IMetadataProvider BuddyMetadataProvider { get; private set;}
        public static IMetadataProvider XmlMetadataProvider { get; private set; }

        static MetadataHandlerRegistration() {
            BuddyMetadataProvider = new BuddyMetadataProvider();
            XmlMetadataProvider = new XmlMetadataProvider();
        }

        public static void Register(Type dataContextType, IMetadataProvider metadataProvider) {
            if (null == metadataProvider) {
                throw new ArgumentNullException("metadataProvider");
            }

            RegisterMetadataHandlerInternal(dataContextType, metadataProvider);
        }

        private static void RegisterMetadataHandlerInternal(Type dataContextType, IMetadataProvider provider) {
            if (dataContextType == null) {
                throw new ArgumentNullException("dataContextType");
            }

            // ensure we have a linq data context type
            Type linqDataContextType = typeof(DataContext);
            if (!linqDataContextType.IsAssignableFrom(dataContextType)) {
                throw new ArgumentException(String.Format("The provided type {0} must iherit from {1}.", dataContextType.FullName, linqDataContextType.FullName), "dataContextType");
            }

            // make sure we havn't registered it before
            if (s_registeredDataContexts.Contains(dataContextType)) {
                throw new InvalidOperationException(String.Format("The data context {0} has already been registered. Each data context can registered only once.", dataContextType.FullName));
            }

            // register providers for all model types that are referenced in the data context type
            foreach (PropertyInfo propertyInfo in dataContextType.GetProperties()) {
                if (IsValidModelProperty(propertyInfo)) {
                    Type modelType = propertyInfo.PropertyType.GetGenericArguments()[0];
                    if (!s_registeredProviders.ContainsKey(modelType)) {
                        List<TypeDescriptionProvider> providers = new List<TypeDescriptionProvider>();
                        
                        TypeDescriptionProvider metadataProvider = new MetadataTypeDescriptionProvider(modelType, provider);
                        TypeDescriptionProvider oldStyleMetadataProvider = new OldStyleMetadataTypeDescriptionProvider(metadataProvider);

                        // not sure if we need to/should register buddyMetadataProvider, since oldStyleMetadataProvider is chained off of it.
                        // does not seem to hurt anything though
                        TypeDescriptor.AddProvider(metadataProvider, modelType);
                        TypeDescriptor.AddProvider(oldStyleMetadataProvider, modelType);

                        providers.Add(oldStyleMetadataProvider);
                        providers.Add(metadataProvider);
                        s_registeredProviders[modelType] = providers;
                    }
                }
            }

            // remember that it's been registered
            s_registeredDataContexts.Add(dataContextType);
        }

        private static bool IsValidModelProperty(PropertyInfo property) {
            if (!property.PropertyType.IsGenericType) {
                return false;
            }

            return property.PropertyType.GetGenericTypeDefinition() == typeof(Table<>);
        }

        internal static void UnregisterAll() {
            foreach (Type type in s_registeredProviders.Keys) {
                foreach (TypeDescriptionProvider provider in s_registeredProviders[type]) {
                    TypeDescriptor.RemoveProvider(provider, type);
                }
            }
            s_registeredProviders.Clear();
            s_registeredDataContexts.Clear();
        }
    }
}
