﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Runtime.Remoting;
using System.Xml;
using System.Globalization;

namespace Microsoft.Web.DynamicData.Metadata {
    internal class XmlMetadataAttributeHandler {
        private Dictionary<string, XmlMetadataAttributeHandlerConfigElement> _configuration = new Dictionary<string, XmlMetadataAttributeHandlerConfigElement>();
        private delegate object ParseFunction(string s);
        private Dictionary<Type, ParseFunction> _parseFunctions = new Dictionary<Type, ParseFunction>();

        public XmlMetadataAttributeHandler(List<XmlMetadataAttributeHandlerConfigElement> config) {
            foreach (XmlMetadataAttributeHandlerConfigElement configElement in config) {
                _configuration[configElement.TagPrefix] = configElement;
            }

            InitializeTypeParseFunctions();
        }

        private void InitializeTypeParseFunctions() {
            // This collection is meant to handle all valid attribute parameter types.
            // See C# Language Specification section 17.1.3

            _parseFunctions[typeof(byte)] = s => byte.Parse(s);
            _parseFunctions[typeof(bool)] = s => bool.Parse(s);
            _parseFunctions[typeof(char)] = s => char.Parse(s);
            // decimal is not supported by the spec
            _parseFunctions[typeof(double)] = s => double.Parse(s);
            _parseFunctions[typeof(float)] = s => float.Parse(s);
            _parseFunctions[typeof(int)] = s => int.Parse(s);
            _parseFunctions[typeof(long)] = s => long.Parse(s);
            _parseFunctions[typeof(sbyte)] = s => sbyte.Parse(s);
            _parseFunctions[typeof(short)] = s => short.Parse(s);
            _parseFunctions[typeof(string)] = s => s;
            _parseFunctions[typeof(uint)] = s => uint.Parse(s);
            _parseFunctions[typeof(ulong)] = s => ulong.Parse(s);
            _parseFunctions[typeof(ushort)] = s => ushort.Parse(s);

            _parseFunctions[typeof(object)] = s => s; // we have no idea what an object property is meant to be so we leave it as a string
            _parseFunctions[typeof(Type)] = s => Type.GetType(s);

            // TODO; single-dimensional arrays
        }

        public BaseMetadataAttribute CreateAttribute(XmlElement metadataXmlElement) {
            if (String.IsNullOrEmpty(metadataXmlElement.Prefix)) {
                throw new InvalidOperationException(String.Format(CultureInfo.CurrentCulture, "Missing tag prefix in attribute tag '{0}'", metadataXmlElement.Name));
            }

            string tagPrefix = metadataXmlElement.Prefix;
            XmlMetadataAttributeHandlerConfigElement configurationElement;
            if (_configuration.TryGetValue(tagPrefix, out configurationElement)) {
                BaseMetadataAttribute attribute = CreateAttribute(metadataXmlElement, configurationElement);
                SetAttributeProperties(metadataXmlElement, attribute);
                return attribute;
            }

            throw new InvalidOperationException(String.Format(CultureInfo.CurrentCulture, "Unknown tag prefix in attribute tag '{0}'", metadataXmlElement.Name));
        }

        private BaseMetadataAttribute CreateAttribute(XmlElement metadataXmlElement, XmlMetadataAttributeHandlerConfigElement configurationElement) {
            try {
                // try *Attribute naming pattern
                string fullName = String.Format(CultureInfo.InvariantCulture, "{0}.{1}Attribute", configurationElement.Namespace, metadataXmlElement.LocalName);
                return TryCreateAttribute(configurationElement.Assembly, fullName);
            }
            catch (TypeLoadException e1) {
                try {
                    // try normal name
                    string fullName = String.Format(CultureInfo.InvariantCulture, "{0}.{1}", configurationElement.Namespace, metadataXmlElement.LocalName);
                    return TryCreateAttribute(configurationElement.Assembly, fullName);
                }
                catch (TypeLoadException) {
                    throw new InvalidOperationException(String.Format(CultureInfo.CurrentCulture, "Can't instantiate attribute class for attribute '{0}'", metadataXmlElement.Name), e1);
                }
            }
        }

        private BaseMetadataAttribute TryCreateAttribute(string assemblyName, string typeName) {
            ObjectHandle handle = Activator.CreateInstance(assemblyName, typeName);
            return (BaseMetadataAttribute)handle.Unwrap();
        }

        private ParseFunction GetParseFunction(Type propertyType) {
            if (propertyType.IsEnum) {
                return s => ParseEnumType(propertyType, s);
            }

            ParseFunction parseFunction = null;
            _parseFunctions.TryGetValue(propertyType, out parseFunction);
            return parseFunction;
        }

        private object ParseEnumType(Type enumType, string value) {
            return Enum.Parse(enumType, value);
        }

        private void SetAttributeProperties(XmlElement metadataXmlElement, BaseMetadataAttribute attribute) {
            Dictionary<string, PropertyInfo> properties = new Dictionary<string, PropertyInfo>();
            foreach (PropertyInfo info in attribute.GetType().GetProperties()) {
                properties[info.Name] = info;
            }

            foreach (XmlAttribute xmlAttribute in metadataXmlElement.Attributes) {
                if (xmlAttribute.Prefix == "xmlns")
                    continue;

                PropertyInfo property;
                if (properties.TryGetValue(xmlAttribute.LocalName, out property)) {
                    Type propertyType = property.PropertyType;
                    ParseFunction parseFunction = GetParseFunction(propertyType);

                    if (null == parseFunction) {
                        throw new InvalidOperationException(String.Format(CultureInfo.CurrentCulture, "Unknown property type {0} matching attribute {1} in element {2}.", propertyType.FullName, xmlAttribute.Name, xmlAttribute.OwnerElement.Name));
                    }

                    object value = parseFunction(metadataXmlElement.GetAttribute(xmlAttribute.LocalName));
                    property.SetValue(attribute, value, null);
                }
                else {
                    throw new InvalidOperationException(String.Format(CultureInfo.CurrentCulture, "The xml attribute '{0}' in element '{1}' has no matching property in the attribute type '{2}'.", xmlAttribute.LocalName, xmlAttribute.OwnerElement.Name, attribute.GetType().FullName));
                }
            }
        }
    }
}
